from timemachines.synthetic import brownian_with_noise, brownian_with_exogenous
from timemachines.skating import prior, residuals
from sklearn.metrics import mean_squared_error, mean_absolute_error, mean_absolute_percentage_error
from scipy.stats import energy_distance
import numpy as np
from timemachines.conventions import targets
from typing import List




# Evaluation of skaters
#   (pretty standard scoring rules)

def evaluate_sklearn_metric(f, ys, metric, n_burn=50, **kwargs):
    """ Useful for a quick test """
    xs = prior(f=f,ys=ys,**kwargs)
    yt = targets(ys)
    return metric(yt[n_burn:], xs[n_burn:])


def evaluate_mean_squared_error(f, ys, n_burn=50, **kwargs):
    return evaluate_sklearn_metric(f=f, ys=ys, metric=mean_squared_error, n_burn=n_burn, **kwargs)


def evaluate_mean_absolute_error(f, ys, n_burn=50, **kwargs):
    return evaluate_sklearn_metric(f=f, ys=ys, metric=mean_absolute_error, n_burn=n_burn, **kwargs)


def evaluate_mean_absolute_percentage_error(f, ys, n_burn=50, **kwargs):
    return evaluate_sklearn_metric(f=f, ys=ys, metric=mean_absolute_percentage_error(), n_burn=n_burn, **kwargs)


def quick_brown_fox(f, n=120, **kwargs):
    """ Useful for a quick test of a skater, w/o exogenous inputs """
    ys = brownian_with_noise(n=n)
    return evaluate_mean_squared_error(f=f,ys=ys, **kwargs)


def lazy_dog(f, n=120, **kwargs):
    """ Useful for a quick test of a skater, w/ exogenous inputs """
    ys = brownian_with_exogenous(n=n)
    return evaluate_mean_squared_error(f=f,ys=ys,**kwargs)


def quick_brown_fox_randomized(f, n=200, **kwargs):
    """ Useful for a quick test of a skater, w/o exogenous inputs """
    r = np.random.rand(1)
    ys = brownian_with_noise(n=n)
    try:
        rmse = evaluate_mean_squared_error(f=f,ys=ys, **kwargs)
    except:
        raise('Error running '+f.__name__ + ' with r='+str(r))


# Energy distance between residuals of consecutive epochs
# (a more speculative way to evaluate point estimates)

def chunk_to_end(l:List, n:int)-> List[List]:
    """
        :param n: Size of batches
    """
    rl = list(reversed(l))
    chunks = [ list(reversed(rl[x:x + n])) for x in range(0, len(rl), n) ]
    return list(reversed(chunks[:-1]))


def evaluate_energy(f, ys=None, k=1, ats=None, ts=None, e=None, r=0.5, n_burn=50, n_epoch=100):
    r = residuals(f=f, ys=ys, k=k, ats=ats, ts=ts, e=e, r=r, n_burn=n_burn)
    r_chunks = chunk_to_end(r,n_epoch)
    return np.mean([ energy_distance(u_values=u_values,v_values=v_values) for u_values, v_values in zip( r_chunks[1:],r_chunks[:-1] )])


def brownian_energy(f, n=500, **kwargs):
    ys = brownian_with_noise(n)
    return evaluate_energy(f=f,ys=ys, **kwargs)


def exogenous_energy(f, n=500, **kwargs):
    ys = brownian_with_exogenous(n)
    return evaluate_energy(f=f,ys=ys, **kwargs)





import matplotlib.pyplot as plt
import numpy as np
from timemachines.synthetic import brownian_with_exogenous, brownian_with_noise
from timemachines.skating import prior


# Quick ways to inspect skaters


def prior_plot(f, ys=None, k=None, ats=None, ts=None, e=None, r=0.5, x0=np.nan, n=150, n_plot=25):
    """
         Apply state machine to univariate series,
         Show observations, and out of sample predictions predictions
    """
    if ys is None:
       ys = brownian_with_noise(n=n)

    if ts is None:
        ts = range(len(ys))

    xs = prior(f=f,ys=ys, k=k, ats=ats, ts=ts, e=e, r=r, x0=x0 )
    ysf = [ [y_] for y_ in ys ]
    plot_observations_and_predictions(ts=ts, xs=xs, ys=ysf, n_plot=n_plot)


def prior_plot_exogenous(f, ys=None, k=None, ats=None, ts=None, e=None, r=0.5, x0=np.nan, n=150, n_plot=25):
    """
          Apply state machine to univariate series,
          Show observations, out of sample predictions predictions, and exogenous variables
     """
    if ys is None:
       ys = brownian_with_exogenous(n)

    if ts is None:
       ts = range(len(ys))

    xs = prior(f=f, ys=ys, k=k, ats=ats, ts=ts, e=e, r=r, x0=x0)
    plot_observations_and_predictions(ts=ts, xs=xs, ys=ys, n_plot=n_plot)


def plot_observations_and_predictions(ts, xs, ys, n_plot:int):
    """
    :param ts:        Time of observation
    :param xs:        Point estimates
    :param ys:        Univariate or Multivariate series of observations
    :param n_plot:    Number of examples to plot (from end)
    :return:
    """
    try:
        ys0 = [y_[0] for y_ in ys]
    except:
        ys0 = ys
    plt.plot(ts[-n_plot:], ys0[-n_plot:], 'b*')
    lgnd = ['target']
    for j in range(1,len(ys[0])):
        ysj = [y_[j] for y_ in ys]
        plt.plot(ts[-n_plot:], ysj[-n_plot:], 'r+')
        lgnd.append('exogenous '+str(j))

    # Run the model
    plt.plot(ts[-n_plot:], xs[-n_plot:], 'g-')
    lgnd.append('prediction')
    plt.legend(lgnd)
    plt.show()





import numpy as np


def brownian_with_noise(n:int)->[float]:
    ys_ = np.cumsum(np.random.randn(n))
    ys = [y + np.random.randn() for y in ys_]
    return list(ys)


def brownian_with_exogenous(n:int)->[(float,)]:
    dys1_ = np.random.randn(n)
    dys2_ = list(0.5*dys1_[2:])+list(np.random.randn(2))
    ys1 = np.cumsum(dys1_) + 0.15*np.random.randn(n)
    ys2 = np.cumsum(dys2_+0.1*dys1_)+0.15*ys1 + 0.15*np.random.randn(n)
    return list(zip(ys1,ys2))





import math
from microconventions.zcurve_conventions import ZCurveConventions
from typing import List, Union, Tuple, Any

# The SKATER convention:
# ----------------------
# A time series model is a function with 7 inputs and 3 outputs
#
#        x, s, w = f( y, s, k, a, t, e, r )
#
# Some additional, optional conventions are expressed in the help functions in this file, but that
# is more for convenient.

# Inputs
Y_TYPE = Union[float, List[float]]  # Observed data, where y[0] is usually assumed to be the 'target'
S_TYPE = Any  # State previously received from callee
K_TYPE = int  # Number of steps ahead to forecast - usually integer
A_TYPE = float  # Known-in advance or other action-conditional variables
# See README.md for discussion of space-filling curves
T_TYPE = Union[float, int]  # Epoch time of observation
E_TYPE = Union[float, int]  # Expiry in seconds (i.e. how long should callee spend computing)
R_TYPE = float  # Hype(r) Pa(r)amete(r)s for the model


# Again, see README.md for discussion of space-filling curves


# Expresses the y convention
def target(y):
    try:
        return y[0]
    except:
        return y


def targets(ys):
    try:
        return [y[0] for y in ys]
    except:
        return [y for y in ys]


# Outputs

X_TYPE = Union[float, None]  # A point estimate, or some other anchor point deemed helpful
# S_TYPE                       Posterior state.
W_TYPE = Any  # (W)hatever else callee chooses to emit, such as a conf interval


def separate_observations(y: Y_TYPE, dim: int):
    """ Usual convention for interpreting y, and checking dimension
    :param s:
    :param y:
    :returns:  y0, exog
    """
    if dim == 1:
        y0 = y
        exog = None
    else:
        y0 = y[0]
        exog = [y[1:]]
    return y0, exog


def dimension(y):
    try:
        return len(y)
    except:
        return 1


def initialize_buffers(s, y: Y_TYPE):
    s['dim'] = dimension(y)
    s['buffer'] = list()  # Target
    if s['dim'] > 1:
        s['exogenous'] = list()  # Exogenous
    s['model'] = None
    s['advance'] = list()  # Variables known in advance
    s['staleness'] = 0
    return s


def update_buffers(s, a: A_TYPE, exog: [float], y0: float):
    # Store "target" and other observations or vars known in advance
    s['buffer'].append(y0)
    if exog is not None:
        s['exogenous'].append(exog[0])
    if a is not None:
        s['advance'].append(a)
    return s


# The remainder of this module establishes space-filling curve conventions that apply to a and r

def positive_log_scale(u, low, high):
    """ Map u in (0,1) to (low,high) """
    assert 0 < low < high
    log_low = math.log(low)
    log_high = math.log(high)
    x = log_low + u * (log_high - log_low)
    return math.exp(x)


def to_log_space_1d(u, low, high):
    """ Approximately logarithmic map, but allows for ranges spanning zero """
    # Avoid singularity at zero

    if 1e-8 < low < high:
        return positive_log_scale(u=u, low=low, high=high)
    elif low < -1e-8 < 1e-8 < high:
        return -positive_log_scale(1 - u, low=-high, high=-low)
    elif -1e-8 < low < 1e-8 < high:
        return positive_log_scale(u=u, low=1e-8, high=high)
    elif low < -1e-8 < high < 1e-8:
        return -positive_log_scale(1 - u, low=1e-8, high=-low)
    else:
        scale = abs(high - low) / 100
        if u < 0.475:
            u1 = 1 - u / 0.475
            return -positive_log_scale(u=u1, low=scale, high=-low)
        elif 0.475 < u < 0.525:
            u2 = 20 * (u - 0.475)
            return -scale + 2 * u2 * scale
        else:
            u3 = (u - 0.525) / 0.525
            return positive_log_scale(u3, low=scale, high=high)


BOUNDS_TYPE = List[Union[Tuple, List]]  # scipy.optimize style bounds [ (low,high), (low, high),... ]


def to_space(p: float, bounds: BOUNDS_TYPE = None, dim: int = 1):
    """ Interprets p as a point in a rectangle in R^2 or R^3

         :param bounds  [ (low,high), (low,high), (low,high) ] defaults to unit cube
         :param dim     Dimension. Only used if bounds are not supplied.

    """
    if bounds is None:
        bounds = [(0, 1) for _ in range(dim)]
    else:
        dim = len(bounds)

    us = reversed(ZCurveConventions().to_cube(zpercentile=p, dim=dim))  # 0 < us[i] < 1
    return [u * (b[1] - b[0]) + b[0] for u, b in zip(us, bounds)]


def from_space(ps: [float], bounds: BOUNDS_TYPE=None):
    """ [ , ]^n -> [0,1] """
    if bounds is None:
        bounds = [(0, 1) for _ in range(len(ps))]
    us = [(pi - b[0]) / (b[1] - b[0]) for pi, b in zip(ps, bounds)]
    for u in us:
        assert 0 <= u <= 1, "bounds are inconsistent with p=" + str(ps)
    return ZCurveConventions().from_cube(list(reversed(us)))


def to_log_space(p:float, bounds: BOUNDS_TYPE):
    """ Interprets p as a point in a rectangle in R^2 or R^3 using Morton space-filling curve

            :param bounds  [ (low,high), (low,high), (low,high) ] defaults to unit cube
            :param dim     Dimension. Only used if bounds are not supplied.

       Very similar to "to_space" but assumes speed varies with logarithm
       """
    assert 0 <= p <= 1
    dim = len(bounds)
    us = list(reversed(ZCurveConventions().to_cube(zpercentile=p, dim=dim)))  # 0 < us[i] < 1
    return [to_log_space_1d(u, low=b[0], high=b[1]) for u, b in zip(us, bounds)]


def to_int_log_space(p: float, bounds: BOUNDS_TYPE):
    """ Interprets p as a point in an integer lattice in R^2 or R^3 using Morton space-filling curve, integers only

            :param bounds  [ (low,high), (low,high), (low,high) ] defaults to unit cube

       Very similar to "to_space" but assumes speed varies with logarithm
       """
    assert 0 <= p <= 1
    prms = to_log_space(p=p, bounds=bounds)
    return [int(prm) for prm in prms]



from typing import Callable
from timemachines.conventions import Y_TYPE, K_TYPE, A_TYPE, T_TYPE, E_TYPE, R_TYPE, targets
import numpy as np


# Basic iteration of state machines implied by skater functions


def posterior(f:Callable,         # "Model" to run forward through observations
             ys:[Y_TYPE],         # Observations
             k:K_TYPE=1,          # Steps ahead to predict, typically integer
             ats:[A_TYPE]=None,   # Data known in advance, or maybe action-conditional
             ts:[T_TYPE]=None,    # Times of observations (epoch seconds)
             e:E_TYPE=None,       # Computation time limit per observation
             r:R_TYPE=0.5):       # Hype(r)-pa(r)amete(r)s
    """ Compute k-step ahead estimates """
    s = None
    ats = [None for _ in ys] if ats is None else ats
    ts = [None for _ in ys] if ts is None else ts
    xs = list()
    for y, t, a in zip(ys, ats, ts):
        x, s, _ = f(y=y, s=s, k=k, a=a, t=t, e=e, r=r)
        xs.append(x)
    return xs


def prior(f, ys, k:int=1, ats=None, ts=None, e=None, r=0.5, x0=np.nan):
    """ Compute k-step ahead estimates aligned for easy comparison against actual """
    # Note k must be integer
    xs = posterior(f=f, ys=ys[:-k], k=k, ats=ats, ts=ts, e=e, r=r)
    return [x0]*k + xs


def residuals(f, ys:[Y_TYPE], k:int, ats=None, ts=None, e=None, r=0.5, x0=np.nan, n_burn=50):
    xs = prior(f=f,ys=ys,k=k,ats=ats,ts=ts,e=e,r=r,x0=x0)
    yt = targets(ys)
    return np.array(xs[-n_burn:])-np.array(yt[-n_burn:])


import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="timemachines",
    version="0.0.18",
    description="Time series models represented as pure functions with SKATER convention.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/microprediction/timemachines",
    author="microprediction",
    author_email="pcotton@intechinvestments.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["timemachines"],
    test_suite='pytest',
    tests_require=['pytest'],
    include_package_data=True,
    install_requires=["wheel","pathlib","numpy>=1.16.5","importlib-metadata>=1.7.0","microconventions>0.5.0","sklearn","divinity","pmdarima","hyperopt","scipy",
                      "poap","pySOT","funcy","pydlm","optuna","deap","ax-platform","sigopt","statsmodels","tdigest","platypus-opt",
                      "pymoo"],
    entry_points={
        "console_scripts": [
            "timemachines=timemachines.__main__:main",
        ]
    },
)



from timemachines.conventions import to_space, Y_TYPE, from_space, dimension
from timemachines.evaluation import evaluate_energy, evaluate_mean_squared_error
from timemachines.optimizers.compendium import OPTIMIZERS, shgo_cube, optuna_cube
from timemachines.skaters.pmd import pmd_auto


def optimize(f, ys:[Y_TYPE],
             evaluator=evaluate_mean_squared_error,
             optimizer=shgo_cube,
             n_trials=100,
             **kwargs)->(float, float):
    """ Returns best r """

    def objective(u:[float]):
        r = from_space(u)
        return evaluator(f=f,ys=ys,r=r,**kwargs)

    r_star = optimizer(objective, n_trials=n_trials, n_dim=dimension(ys[0]))
    evaluation = objective(r_star)
    return r_star, evaluation


if __name__=='__main__':
    # Example that could take a while
    from timemachines.synthetic import brownian_with_exogenous
    r_star, evaluation = optimize(f=pmd_auto,ys=brownian_with_exogenous(n=300),
                      n_trials=25, optimizer=optuna_cube)
    print("Best hyper-param is "+str(r_star))
    from timemachines.skaters.pmd import pmd_hyperparams
    s = pmd_hyperparams(s=dict(),r=r_star)
    from pprint import pprint
    pprint(s)


